@@ -4,7 +4,7 @@
  * Copyright (C) 2003 Red Hat, Inc.
  *
  * Licensed under the Academic Free License version 2.1
- * 
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
@@ -14,10 +14,10 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
  */
 
@@ -36,28 +36,13 @@
 static DBusList *clients = NULL;
 static DBusLoop *client_loop = NULL;
 
-static dbus_bool_t
-client_watch_callback (DBusWatch     *watch,
-                       unsigned int   condition,
-                       void          *data)
-{
-  /* FIXME this can be done in dbus-mainloop.c
-   * if the code in activation.c for the babysitter
-   * watch handler is fixed.
-   */
- 
-  return dbus_watch_handle (watch, condition);
-}
-
 static dbus_bool_t
 add_client_watch (DBusWatch      *watch,
                   void           *data)
 {
   DBusConnection *connection = data;
-  
-  return _dbus_loop_add_watch (client_loop,
-                               watch, client_watch_callback, connection,
-                               NULL);
+
+  return _dbus_loop_add_watch (client_loop, watch);
 }
 
 static void
@@ -65,23 +50,8 @@ remove_client_watch (DBusWatch      *watch,
                      void           *data)
 {
   DBusConnection *connection = data;
-  
-  _dbus_loop_remove_watch (client_loop,
-                           watch, client_watch_callback, connection);
-}
 
-static void
-client_timeout_callback (DBusTimeout   *timeout,
-                         void          *data)
-{
-  DBusConnection *connection = data;
-
-  dbus_connection_ref (connection);
-
-  /* can return FALSE on OOM but we just let it fire again later */
-  dbus_timeout_handle (timeout);
-
-  dbus_connection_unref (connection);
+  _dbus_loop_remove_watch (client_loop, watch);
 }
 
 static dbus_bool_t
@@ -89,8 +59,8 @@ add_client_timeout (DBusTimeout    *timeout,
                     void           *data)
 {
   DBusConnection *connection = data;
-  
-  return _dbus_loop_add_timeout (client_loop, timeout, client_timeout_callback, connection, NULL);
+
+  return _dbus_loop_add_timeout (client_loop, timeout);
 }
 
 static void
@@ -98,8 +68,8 @@ remove_client_timeout (DBusTimeout    *timeout,
                        void           *data)
 {
   DBusConnection *connection = data;
-  
-  _dbus_loop_remove_timeout (client_loop, timeout, client_timeout_callback, connection);
+
+  _dbus_loop_remove_timeout (client_loop, timeout);
 }
 
 static DBusHandlerResult
@@ -111,27 +81,27 @@ client_disconnect_filter (DBusConnection     *connection,
                                DBUS_INTERFACE_LOCAL,
                                "Disconnected"))
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-    
+
   _dbus_verbose ("Removing client %p in disconnect handler\n",
                  connection);
-  
+
   _dbus_list_remove (&clients, connection);
 
   dbus_connection_unref (connection);
-  
+
   if (clients == NULL)
     {
       _dbus_loop_unref (client_loop);
       client_loop = NULL;
     }
-  
+
   return DBUS_HANDLER_RESULT_HANDLED;
 }
 
 dbus_bool_t
 bus_setup_debug_client (DBusConnection *connection)
 {
-  dbus_bool_t retval;  
+  dbus_bool_t retval;
 
   if (!dbus_connection_add_filter (connection,
                                    client_disconnect_filter,
@@ -146,7 +116,7 @@ bus_setup_debug_client (DBusConnection *connection)
       if (client_loop == NULL)
         goto out;
     }
-  
+
   if (!dbus_connection_set_watch_functions (connection,
                                             add_client_watch,
                                             remove_client_watch,
@@ -154,7 +124,7 @@ bus_setup_debug_client (DBusConnection *connection)
                                             connection,
                                             NULL))
     goto out;
-      
+
   if (!dbus_connection_set_timeout_functions (connection,
                                               add_client_timeout,
                                               remove_client_timeout,
@@ -164,16 +134,16 @@ bus_setup_debug_client (DBusConnection *connection)
 
   if (!_dbus_list_append (&clients, connection))
     goto out;
-  
+
   retval = TRUE;
-  
+
  out:
   if (!retval)
     {
       dbus_connection_remove_filter (connection,
                                      client_disconnect_filter,
                                      NULL);
-      
+
       dbus_connection_set_watch_functions (connection,
                                            NULL, NULL, NULL, NULL, NULL);
       dbus_connection_set_timeout_functions (connection,
@@ -187,7 +157,7 @@ bus_setup_debug_client (DBusConnection *connection)
           client_loop = NULL;
         }
     }
-      
+
   return retval;
 }
 
@@ -196,7 +166,7 @@ bus_test_clients_foreach (BusConnectionForeachFunction  function,
                           void                         *data)
 {
   DBusList *link;
-  
+
   link = _dbus_list_get_first_link (&clients);
   while (link != NULL)
     {
@@ -205,7 +175,7 @@ bus_test_clients_foreach (BusConnectionForeachFunction  function,
 
       if (!(* function) (connection, data))
         break;
-      
+
       link = next;
     }
 }
@@ -214,7 +184,7 @@ dbus_bool_t
 bus_test_client_listed (DBusConnection *connection)
 {
   DBusList *link;
-  
+
   link = _dbus_list_get_first_link (&clients);
   while (link != NULL)
     {
@@ -223,7 +193,7 @@ bus_test_client_listed (DBusConnection *connection)
 
       if (c == connection)
         return TRUE;
-      
+
       link = next;
     }
 
@@ -232,17 +202,17 @@ bus_test_client_listed (DBusConnection *connection)
 
 void
 bus_test_run_clients_loop (dbus_bool_t block_once)
-{  
+{
   if (client_loop == NULL)
     return;
 
   _dbus_verbose ("---> Dispatching on \"client side\"\n");
-  
+
   /* dispatch before we block so pending dispatches
    * won't make our block return early
    */
   _dbus_loop_dispatch (client_loop);
-  
+
   /* Do one blocking wait, since we're expecting data */
   if (block_once)
     {
@@ -262,12 +232,12 @@ bus_test_run_bus_loop (BusContext *context,
                        dbus_bool_t block_once)
 {
   _dbus_verbose ("---> Dispatching on \"server side\"\n");
-  
+
   /* dispatch before we block so pending dispatches
    * won't make our block return early
    */
   _dbus_loop_dispatch (bus_context_get_loop (context));
-  
+
   /* Do one blocking wait, since we're expecting data */
   if (block_once)
     {
@@ -298,7 +268,7 @@ bus_context_new_test (const DBusString *test_data_dir,
   DBusString config_file;
   DBusString relative;
   BusContext *context;
-  
+
   if (!_dbus_string_init (&config_file))
     {
       _dbus_warn ("No memory\n");
@@ -321,25 +291,25 @@ bus_context_new_test (const DBusString *test_data_dir,
       _dbus_string_free (&config_file);
       return NULL;
     }
-  
+
   dbus_error_init (&error);
-  context = bus_context_new (&config_file, FALSE, NULL, NULL, &error);
+  context = bus_context_new (&config_file, FALSE, NULL, NULL, NULL, FALSE, &error);
   if (context == NULL)
     {
       _DBUS_ASSERT_ERROR_IS_SET (&error);
-      
+
       _dbus_warn ("Failed to create debug bus context from configuration file %s: %s\n",
                   filename, error.message);
 
       dbus_error_free (&error);
-      
+
       _dbus_string_free (&config_file);
-      
+
       return NULL;
     }
 
   _dbus_string_free (&config_file);
-  
+
   return context;
 }
 
